@@ -31,7 +31,7 @@
 
 #include "config.h"
 
-#define DEFAULT_DELAY 10
+#define DEFAULT_DELAY "10"
 #define DEFAULT_TIMES -1
 
 #define STDIN_FD 0
@@ -60,12 +60,17 @@ static struct option long_options[] =
 };
 
 typedef struct pump_t {
+    struct pollfd *rpollfd;
+    struct pollfd *wpollfd;
     void *base;
     size_t len;
     off_t offset;
+    int rfd;
+    int wfd;
     int read_closed:1;
     int write_closed:1;
     int exit_on_close:1;
+    int no_close:1;
     int send_eof:1;
 } pump_t;
 
@@ -107,7 +112,9 @@ static int help(const char *name, const char *msg, int code)
             "\n"
             "OPTIONS\n"
             "  -d seconds, --delay=seconds  The number of seconds to back off\n"
-            "    after each attempt.\n"
+            "    after each attempt. Multiple comma separated\n"
+            "    delays allow subsequent delays to be different,\n"
+            "    with the last delay repeated.\n"
             "\n"
             "  -m message, --message=message  A message to include in the notification\n"
             "    when repeat has backed off. Defaults to the\n"
@@ -144,7 +151,7 @@ static int help(const char *name, const char *msg, int code)
             "EXAMPLES\n"
             "  In this basic example, we repeat the command forever.\n"
             "\n"
-            "\t~$ retry --until=success false\n"
+            "\t~$ retry --until=success -- false\n"
             "\tretry: 'false' returned 1, backing off for 10 seconds and trying again...\n"
             "\tretry: 'false' returned 1, backing off for 10 seconds and trying again...\n"
             "\tretry: 'false' returned 1, backing off for 10 seconds and trying again...\n"
@@ -155,11 +162,27 @@ static int help(const char *name, const char *msg, int code)
             "  passed once and once only to the next element in the\n"
             "  pipeline.\n"
             "\n"
-            "\t~$ retry curl --fail http://localhost/entities | \\\\ \n"
+            "\t~$ retry -- curl --fail http://localhost/entities | \\\\ \n"
             "\tjq ... | \\\\ \n"
-            "\tretry curl --fail -X POST http://localhost/resource | \\\\ \n"
+            "\tretry -- curl --fail -X POST http://localhost/resource | \\\\ \n"
             "\tlogger -t resource-init\n"
             "\n"
+            "  In this example, we stagger each delay exponentially\n"
+            "  until 64 seconds, which is then repeated until\n"
+            "  interrupted.\n"
+            "\n"
+            "\t~$ retry --until=success --delay \"1,2,4,8,16,32,64\" -- false\n"
+            "\tretry: false returned 1, backing off for 1 second and trying again...\n"
+            "\tretry: false returned 1, backing off for 2 seconds and trying again...\n"
+            "\tretry: false returned 1, backing off for 4 seconds and trying again...\n"
+            "\tretry: false returned 1, backing off for 8 seconds and trying again...\n"
+            "\tretry: false returned 1, backing off for 16 seconds and trying again...\n"
+            "\tretry: false returned 1, backing off for 32 seconds and trying again...\n"
+            "\tretry: false returned 1, backing off for 64 seconds and trying again...\n"
+            "\tretry: false returned 1, backing off for 64 seconds and trying again...\n"
+            "\tretry: false returned 1, backing off for 64 seconds and trying again...\n"
+            "\t^C\n"
+            "\n"
             "AUTHOR\n"
             "  Graham Leggett <minfrin@sharp.fm>\n"
             "", msg ? msg : "", n, n);
@@ -232,38 +255,44 @@ static int status_match(int status, const char *criteria)
     return 1;
 }
 
-static int pump(const char *name, pump_t *pumps, struct pollfd *fds)
+static int pump(const char *name, pump_t *pumps)
 {
-    nfds_t nfds = PUMPS * 2;
     int i;
 
     while (1) {
-        int stay = 0;
+        nfds_t nfds = 0;
+        struct pollfd fds[PUMPS * 2] = { 0 };
 
         for (i = 0; i < PUMPS; i++) {
 
-            fds[OFFSET(i, READ_FD)].events = 0;
-            fds[OFFSET(i, WRITE_FD)].events = 0;
+            pumps[i].rpollfd = NULL;
+            pumps[i].wpollfd = NULL;
 
             if (!pumps[i].read_closed) {
-                fds[OFFSET(i, READ_FD)].events = POLLIN;
-                stay = 1;
+                pumps[i].rpollfd = fds + nfds;
+                fds[nfds].events = POLLIN;
+                fds[nfds].fd = pumps[i].rfd;
+                fds[nfds].revents = 0;
+                nfds++;
             }
 
             if (!pumps[i].write_closed
                     && (pumps[i].send_eof || pumps[i].len > pumps[i].offset)) {
-                fds[OFFSET(i, WRITE_FD)].events = POLLOUT;
-                stay = 1;
+                pumps[i].wpollfd = fds + nfds;
+                fds[nfds].events = POLLOUT;
+                fds[nfds].fd = pumps[i].wfd;
+                fds[nfds].revents = 0;
+                nfds++;
             }
 
             if (pumps[i].read_closed && pumps[i].exit_on_close) {
-                stay = 0;
+                nfds = 0;
                 break;
             }
 
         }
 
-        if (!stay) {
+        if (!nfds) {
             break;
         }
 
@@ -277,80 +306,91 @@ static int pump(const char *name, pump_t *pumps, struct pollfd *fds)
 
         for (i = 0; i < PUMPS; i++) {
 
-            if (fds[OFFSET(i, READ_FD)].revents & POLLIN) {
-                void *base;
-                ssize_t num;
+            if (pumps[i].rpollfd) {
 
-                base = realloc(pumps[i].base, pumps[i].len + BUFFER_SIZE);
-                if (!base) {
+                if (pumps[i].rpollfd->revents & POLLIN) {
+                    void *base;
+                    ssize_t num;
 
-                    fprintf(stderr, "%s: Out of memory, giving up.\n", name);
+                    base = realloc(pumps[i].base, pumps[i].len + BUFFER_SIZE);
+                    if (!base) {
+
+                        fprintf(stderr, "%s: Out of memory, giving up.\n",
+                                name);
+
+                        return -1;
+                    }
+                    pumps[i].base = base;
+
+                    num = read(pumps[i].rfd, base + pumps[i].len, BUFFER_SIZE);
+
+                    if (num < 0) {
+
+                        fprintf(stderr, "%s: Could not read, giving up: %s\n",
+                                name, strerror(errno));
+
+                        return -1;
+                    } else if (num == 0) {
+                        pumps[i].read_closed = 1;
+                        pumps[i].send_eof = 1;
+                        close(pumps[i].rfd);
+                    } else {
+                        pumps[i].len += num;
+                    }
 
-                    return -1;
                 }
-                pumps[i].base = base;
 
-                num = read(fds[OFFSET(i, READ_FD)].fd, base + pumps[i].len, BUFFER_SIZE);
+                if ((pumps[i].rpollfd->revents & POLLHUP)
+                        || (pumps[i].rpollfd->revents & POLLERR)
+                        || (pumps[i].rpollfd->revents & POLLNVAL)) {
 
-                if (num < 0) {
-
-                    fprintf(stderr, "%s: Could not read, giving up: %s\n", name,
-                            strerror(errno));
-
-                    return -1;
-                }
-                else if (num == 0) {
                     pumps[i].read_closed = 1;
                     pumps[i].send_eof = 1;
-                }
-                else {
-                    pumps[i].len += num;
+                    close(pumps[i].rfd);
+
                 }
 
             }
 
-            if ((fds[OFFSET(i, READ_FD)].revents & POLLHUP) ||
-                    (fds[OFFSET(i, READ_FD)].revents & POLLERR) ||
-                    (fds[OFFSET(i, READ_FD)].revents & POLLNVAL)) {
+            if (pumps[i].wpollfd) {
 
-                pumps[i].read_closed = 1;
-                pumps[i].send_eof = 1;
+                if (pumps[i].wpollfd->revents & POLLOUT) {
+                    ssize_t num;
 
-            }
+                    num = write(pumps[i].wfd, pumps[i].base + pumps[i].offset,
+                            pumps[i].len - pumps[i].offset);
 
-            if (fds[OFFSET(i, WRITE_FD)].revents & POLLOUT) {
-                ssize_t num;
+                    if (num < 0) {
 
-                num = write(fds[OFFSET(i, WRITE_FD)].fd,
-                        pumps[i].base + pumps[i].offset,
-                        pumps[i].len - pumps[i].offset);
+                        fprintf(stderr, "%s: Could not write, giving up: %s\n",
+                                name, strerror(errno));
 
-                if (num < 0) {
+                        return -1;
+                    } else {
+                        pumps[i].offset += num;
+                    }
 
-                    fprintf(stderr, "%s: Could not write, giving up: %s\n", name,
-                            strerror(errno));
+                    if (pumps[i].read_closed
+                            && pumps[i].offset == pumps[i].len) {
+
+                        pumps[i].write_closed = 1;
+                        close(pumps[i].wfd);
+
+                    }
 
-                    return -1;
-                }
-                else {
-                    pumps[i].offset += num;
                 }
 
-                if (pumps[i].read_closed && pumps[i].offset == pumps[i].len) {
+                if ((pumps[i].wpollfd->revents & POLLHUP)
+                        || (pumps[i].wpollfd->revents & POLLERR)
+                        || (pumps[i].wpollfd->revents & POLLNVAL)) {
 
                     pumps[i].write_closed = 1;
+                    close(pumps[i].wfd);
 
                 }
 
             }
 
-            if ((fds[OFFSET(i, WRITE_FD)].revents & POLLERR) ||
-                    (fds[OFFSET(i, WRITE_FD)].revents & POLLNVAL)) {
-
-                pumps[i].write_closed = 1;
-
-            }
-
         }
 
     }
@@ -364,23 +404,29 @@ int main (int argc, char **argv)
     const char *repeat_until = "success";
     const char *repeat_while = NULL;
     const char *message = NULL;
-    char *endptr = NULL;
+    char *delay = DEFAULT_DELAY;
     pump_t pumps[PUMPS] = { 0 };
     int c, status = 0, i;
-    long int delay = DEFAULT_DELAY;
+    long int d = 0;
     long int times = DEFAULT_TIMES;
 
-    while ((c = getopt_long(argc, argv, "d:m:t:u:w:hv", long_options, NULL)) != -1) {
+    while ((c = getopt_long(argc, argv, "+d:m:t:u:w:hv", long_options, NULL)) != -1) {
 
         switch (c)
         {
         case 'd':
-            errno = 0;
-            delay = strtol(optarg, &endptr, 10);
+            delay = optarg;
 
-            if (errno || endptr[0] || delay < 0) {
-                return help(name, "Delay must be bigger or equal to 0.\n", EXIT_FAILURE);
-            }
+            do {
+                errno = 0;
+
+                d = strtol(optarg, &optarg, 10);
+
+                if (errno || (optarg[0] && optarg[0] != ',') || d < 0) {
+                    return help(name, "Delay(s) must be bigger or equal to 0.\n", EXIT_FAILURE);
+                }
+
+            } while (optarg++[0] == ',');
 
             break;
         case 'm':
@@ -389,9 +435,9 @@ int main (int argc, char **argv)
             break;
         case 't':
             errno = 0;
-            times = strtol(optarg, &endptr, 10);
+            times = strtol(optarg, &optarg, 10);
 
-            if (errno || endptr[0] || times < -1) {
+            if (errno || optarg[0] || times < -1) {
                 return help(name, "Times must be bigger or equal to -1.\n", EXIT_FAILURE);
             }
 
@@ -476,18 +522,17 @@ int main (int argc, char **argv)
 
         /* parent */
         else {
-            struct pollfd fds[PUMPS * 2];
 
             /* handle stdin */
 
-            fds[OFFSET(STDIN_FD, READ_FD)].fd = STDIN_FD;
-            fds[OFFSET(STDIN_FD, WRITE_FD)].fd = inpair[WRITE_FD];
+            pumps[STDIN_FD].rfd = dup(STDIN_FD);
+            pumps[STDIN_FD].wfd = inpair[WRITE_FD];
             close(inpair[READ_FD]);
 
             /* handle stdout */
 
-            fds[OFFSET(STDOUT_FD, READ_FD)].fd = outpair[READ_FD];
-            fds[OFFSET(STDOUT_FD, WRITE_FD)].fd = STDOUT_FD;
+            pumps[STDOUT_FD].rfd = outpair[READ_FD];
+            pumps[STDOUT_FD].wfd = dup(STDOUT_FD);
             close(outpair[WRITE_FD]);
 
             /* prevent write to stdout, we will handle it later */
@@ -497,16 +542,19 @@ int main (int argc, char **argv)
             pumps[STDOUT_FD].exit_on_close = 1;
 
             /* pump all data */
-            if (pump(name, pumps, fds)) {
+            if (pump(name, pumps)) {
                 status = EXIT_FAILURE;
                 break;
             }
 
             close(inpair[WRITE_FD]);
             close(outpair[READ_FD]);
+            close(pumps[STDIN_FD].rfd);
+            close(pumps[STDOUT_FD].wfd);
 
             /* reset stdin in case we repeat the command */
             pumps[STDIN_FD].offset = 0;
+            pumps[STDIN_FD].write_closed = 0;
 
             /* wait for the child process to be done */
             do {
@@ -543,17 +591,28 @@ int main (int argc, char **argv)
                 /* failure - write stdout to stderr */
                 fwrite(pumps[STDOUT_FD].base, pumps[STDOUT_FD].len, 1, stderr);
 
-                if (delay) {
-                    fprintf(stderr,
-                            "%s: '%s' returned %d, backing off for %ld second%s and trying again...\n",
-                            name, message ? message : argv[optind], status,
-                            delay, delay > 1 ? "s" : "");
+                /* reset stdout for a go-around */
+                free(pumps[STDOUT_FD].base);
+                memset(&pumps[STDOUT_FD], 0, sizeof(pump_t));
 
-                    sleep(delay);
+                if (delay[0]) {
+                    d = strtol(delay, &delay, 10);
+                }
+                if (delay[0] == ',') {
+                    delay++;
+                }
+
+                if (d) {
+                    fprintf(stderr,
+                            "%s: %s returned %d, backing off for %ld second%s and trying again...\n",
+                            name, message ? message : argv[optind], status,
+                            d, d > 1 ? "s" : "");
+
+                    sleep(d);
                 }
                 else {
                     fprintf(stderr,
-                            "%s: '%s' returned %d, trying again...\n",
+                            "%s: %s returned %d, trying again...\n",
                             name, message ? message : argv[optind], status);
                 }
 
